from pathlib import Path

from jinja2 import Template

from truss.base import truss_config
from truss.cli.train.deploy_checkpoints.deploy_checkpoints_helpers import (
    START_COMMAND_ENVVAR_NAME,
)
from truss.cli.train.types import DeployCheckpointsConfigComplete
from truss_train.definitions import FullCheckpoint

from .deploy_checkpoints_helpers import (
    setup_base_truss_config,
    setup_environment_variables_and_secrets,
)

VLLM_FULL_START_COMMAND = Template(
    'sh -c "{%if envvars %}{{ envvars }} {% endif %}vllm serve {{ model_path }}'
    + " --port 8000"
    + " --tensor-parallel-size {{ specify_tensor_parallelism }}"
    + " --dtype bfloat16"
    + '"'
)


def render_vllm_full_truss_config(
    checkpoint_deploy: DeployCheckpointsConfigComplete,
) -> truss_config.TrussConfig:
    """Render truss config specifically for full checkpoints using vLLM."""
    truss_deploy_config = setup_base_truss_config(checkpoint_deploy)

    start_command_envvars = setup_environment_variables_and_secrets(
        truss_deploy_config, checkpoint_deploy
    )

    checkpoint_str = _build_full_checkpoint_string(truss_deploy_config)

    accelerator = checkpoint_deploy.compute.accelerator

    start_command_args = {
        "model_path": checkpoint_str,
        "envvars": start_command_envvars,
        "specify_tensor_parallelism": accelerator.count if accelerator else 1,
    }
    # Note: we set the start command as an environment variable in supervisord config.
    # This is so that we don't have to change the supervisord config when the start command changes.
    # Our goal is to reduce the number of times we need to rebuild the image, and allow us to deploy faster.
    start_command = VLLM_FULL_START_COMMAND.render(**start_command_args)
    truss_deploy_config.environment_variables[START_COMMAND_ENVVAR_NAME] = start_command
    # Note: supervisord uses the convention %(ENV_VAR_NAME)s to access environment variable VAR_NAME
    truss_deploy_config.docker_server.start_command = (  # type: ignore[union-attr]
        f"%(ENV_{START_COMMAND_ENVVAR_NAME})s"
    )

    return truss_deploy_config


def hydrate_full_checkpoint(
    job_id: str, checkpoint_id: str, checkpoint: dict
) -> FullCheckpoint:
    """Create a Checkpoint object for full model weights."""
    # NOTE: Slash at the end is important since it means the checkpoint is a directory
    paths = [f"rank-0/{checkpoint_id}/"]
    return FullCheckpoint(training_job_id=job_id, paths=paths)


def _build_full_checkpoint_string(truss_deploy_config) -> str:
    """Build checkpoint string from artifact references for full checkpoints.

    Args:
        truss_deploy_config: The truss deploy configuration containing training checkpoints.

    Returns:
        A space-separated string of checkpoint paths.
    """
    checkpoint_parts = []
    for (
        truss_checkpoint
    ) in truss_deploy_config.training_checkpoints.artifact_references:  # type: ignore
        ckpt_path = Path(
            truss_deploy_config.training_checkpoints.download_folder,  # type: ignore
            truss_checkpoint.training_job_id,
            truss_checkpoint.paths[0],
        )
        checkpoint_parts.append(str(ckpt_path))

    return " ".join(checkpoint_parts)



import inspect
from pathlib import Path
from typing import Any

from truss.base.truss_spec import TrussSpec
from truss.contexts.local_loader.truss_module_loader import truss_module_loaded
from truss.contexts.local_loader.utils import (
    prepare_secrets,
    signature_accepts_keyword_arg,
)
from truss.contexts.truss_context import TrussContext
from truss.templates.server.common.patches import apply_patches


class LoadModelLocal(TrussContext):
    """Loads a Truss model locally.

    The loaded model can be used to make predictions for quick testing.
    Runs in the current pip environment directly. Assumes all requirements and
    system packages are already installed.
    """

    @staticmethod
    def run(truss_dir: Path):
        spec = TrussSpec(truss_dir)
        with truss_module_loaded(
            str(truss_dir),
            spec.model_module_fullname,
            spec.bundled_packages_dir.name,
            [str(path.resolve()) for path in spec.external_package_dirs_paths],
        ) as module:
            model_class = getattr(module, spec.model_class_name)
            model_class_signature = inspect.signature(model_class)
            model_init_params: dict[str, Any] = {}
            if signature_accepts_keyword_arg(model_class_signature, "config"):
                model_init_params["config"] = spec.config.to_dict()
            if signature_accepts_keyword_arg(model_class_signature, "data_dir"):
                model_init_params["data_dir"] = truss_dir / "data"
            if signature_accepts_keyword_arg(model_class_signature, "secrets"):
                model_init_params["secrets"] = prepare_secrets(spec)
            apply_patches(spec.apply_library_patches, spec.requirements)
            model = model_class(**model_init_params)
            if hasattr(model, "load"):
                model.load()
            return model



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from fastapi import FastAPI, WebSocket
from httpx import AsyncClient, Response
from httpx_ws import AsyncWebSocketSession
from httpx_ws import _exceptions as httpx_ws_exceptions
from wsproto.events import BytesMessage, TextMessage

from truss.templates.control.control.endpoints import proxy_ws


@pytest.fixture
def app():
    app = FastAPI()
    app.state.proxy_client = AsyncClient(base_url="http://localhost:8080")
    app.state.logger = MagicMock()
    return app


@pytest.fixture
def client_ws(app):
    ws = WebSocket({"type": "websocket", "app": app}, None, None)  # type: ignore
    ws.receive = AsyncMock()
    ws.send_text = AsyncMock()
    ws.send_bytes = AsyncMock()
    ws.accept = AsyncMock()
    ws.close = AsyncMock()
    return ws


@pytest.mark.asyncio
async def test_proxy_ws_bidirectional_messaging(client_ws):
    """Test that both directions of communication work and clean up properly"""
    client_ws.receive.side_effect = [
        {"type": "websocket.receive", "text": "msg1"},
        {"type": "websocket.receive", "text": "msg2"},
        {"type": "websocket.disconnect"},
    ]

    mock_server_ws = AsyncMock(spec=AsyncWebSocketSession)
    mock_server_ws.receive.side_effect = [
        TextMessage(data="response1"),
        TextMessage(data="response2"),
        None,  # server closing connection
    ]
    mock_server_ws.__aenter__.return_value = mock_server_ws
    mock_server_ws.__aexit__.return_value = None

    with patch(
        "truss.templates.control.control.endpoints.aconnect_ws",
        return_value=mock_server_ws,
    ):
        await proxy_ws(client_ws)

    assert mock_server_ws.send_text.call_count == 2
    assert mock_server_ws.send_text.call_args_list == [(("msg1",),), (("msg2",),)]
    assert client_ws.send_text.call_count == 2
    assert client_ws.send_text.call_args_list == [(("response1",),), (("response2",),)]
    client_ws.close.assert_called_once()


@pytest.mark.asyncio
async def test_proxy_ws_binary_message(client_ws):
    test_bytes = b"binary data"
    client_ws.receive.side_effect = [
        {"type": "websocket.receive", "bytes": test_bytes},
        {"type": "websocket.disconnect"},
    ]

    mock_server_ws = AsyncMock(spec=AsyncWebSocketSession)
    mock_server_ws.receive.side_effect = [
        BytesMessage(data=b"binary response"),
        None,  # server closing connection
    ]
    mock_server_ws.__aenter__.return_value = mock_server_ws
    mock_server_ws.__aexit__.return_value = None

    with patch(
        "truss.templates.control.control.endpoints.aconnect_ws",
        return_value=mock_server_ws,
    ):
        await proxy_ws(client_ws)

    # Verify interactions
    client_ws.accept.assert_called_once()
    mock_server_ws.send_bytes.assert_called_once_with(test_bytes)
    client_ws.send_bytes.assert_called_once_with(b"binary response")
    client_ws.close.assert_called_once()


@pytest.mark.asyncio
async def test_proxy_ws_closes_client_upon_server_connection_error(client_ws):
    with patch(
        "truss.templates.control.control.endpoints.aconnect_ws",
        side_effect=httpx_ws_exceptions.WebSocketUpgradeError(
            Response(status_code=503)
        ),
    ):
        await proxy_ws(client_ws)

    client_ws.close.assert_called_once()
    assert client_ws.app.state.logger.warning.called
